@@ -9,7 +9,7 @@ use x11rb::protocol::xproto::ConnectionExt as _;
 use x11rb::xcb_ffi::XCBConnection;
 
 use crate::dialog::Dialog;
-use crate::errors::{Result, Unsupported, X11ErrorString as _};
+use crate::errors::{Result, Unsupported};
 use crate::{Connection, XId};
 
 // Hide u32 because CompletionNotify events might not come in in order or the serial might have
@@ -24,7 +24,6 @@ pub struct Backbuffer<'a> {
     serial: u32,
     vsync_completed: bool,
     backbuffer_dirty: bool,
-    update_pending: bool,
     backbuffer_idle: bool,
     surface: XcbSurface<'a>,
     pub(super) cr: cairo::Context,
@@ -42,9 +41,9 @@ pub struct Cookie<'a> {
 
 impl<'a> Cookie<'a> {
     pub fn reply(self) -> Result<Backbuffer<'a>> {
-        let version = self.version.reply().map_xerr()?;
+        let version = self.version.reply()?;
         if let Some(caps) = self.caps {
-            let caps = caps.reply().map_xerr()?;
+            let caps = caps.reply()?;
             debug!(
                 "present version: {}.{}, capabilities: async {}, fence: {}",
                 version.major_version,
@@ -60,7 +59,6 @@ impl<'a> Cookie<'a> {
             eid: None,
             serial: 0,
             vsync_completed: true,
-            update_pending: false,
             backbuffer_dirty: true,
             backbuffer_idle: true,
             surface: self.surface,
@@ -103,7 +101,7 @@ impl<'a> Backbuffer<'a> {
     }
 
     pub fn init(&mut self, frontbuffer: xproto::Window, dialog: &mut Dialog) -> Result<()> {
-        self.eid = Some(self.conn.generate_id().map_xerr()?);
+        self.eid = Some(self.conn.generate_id()?);
         self.conn.present_select_input(
             self.eid.unwrap(),
             frontbuffer,
@@ -125,39 +123,28 @@ impl<'a> Backbuffer<'a> {
             if let Some((width, height)) = self.resize_requested {
                 trace!("resize requested");
                 let surface_cleared = self.surface.resize(width, height)?;
-                dialog.resize(
-                    &self.cr,
-                    width,
-                    height,
-                    FrameId(self.get_next_serial()),
-                    surface_cleared,
-                );
+                dialog.resize(&self.cr, width, height, surface_cleared);
                 self.resize_requested = None;
             } else {
-                dialog.update(&self.cr, FrameId(self.get_next_serial()));
+                dialog.repaint(&self.cr);
             }
             self.surface.flush();
-            self.update_pending = false;
+            dialog.set_painted(FrameId(self.get_next_serial()));
             self.present()?;
             Ok(true)
         } else {
             trace!("update: backbuffer not idle");
-            self.update_pending = true;
             Ok(false)
         }
     }
 
-    pub fn on_idle_notify(&mut self, ev: &present::IdleNotifyEvent) -> bool {
+    pub fn on_idle_notify(&mut self, ev: &present::IdleNotifyEvent) {
         if ev.serial == self.serial {
             self.backbuffer_idle = true;
             trace!("idle notify: backbuffer became idle");
-            if self.update_pending {
-                return true;
-            }
         } else {
             trace!("idle notify: not idle");
         }
-        false
     }
 
     pub fn on_vsync_completed(&mut self, ev: present::CompleteNotifyEvent) -> FrameId {
@@ -234,7 +221,7 @@ impl<'a> XcbSurface<'a> {
         width: u16,
         height: u16,
     ) -> Result<Self> {
-        let pixmap = conn.generate_id().map_xerr()?;
+        let pixmap = conn.generate_id()?;
         conn.create_pixmap(depth, pixmap, drawable, width, height)?;
 
         let surface = Self::create(conn, pixmap, visual_type, width, height);
@@ -263,6 +250,7 @@ impl<'a> XcbSurface<'a> {
         let cairo_visual =
             unsafe { cairo::XCBVisualType::from_raw_none(&mut xcb_visualtype as *mut _ as _) };
         let cairo_drawable = cairo::XCBDrawable(drawable);
+        trace!("creating cairo::XCBSurface {}, {}", width, height);
         cairo::XCBSurface::create(
             &cairo_conn,
             &cairo_drawable,
@@ -303,7 +291,7 @@ impl<'a> XcbSurface<'a> {
         new_width: u16,
         new_height: u16,
     ) -> Result<()> {
-        let pixmap = self.conn.generate_id().map_xerr()?;
+        let pixmap = self.conn.generate_id()?;
         self.conn
             .create_pixmap(self.depth, pixmap, drawable, new_width, new_height)?;
 
