@@ -1,9 +1,11 @@
+use std::io::Write;
 use std::path::Path;
 
 use serde::{Deserialize, Deserializer, Serialize, Serializer};
 use toml::Value;
 
-use crate::errors::{bail, Context as _, Result};
+use crate::bail;
+use crate::errors::{Context as _, Error, Result};
 
 pub const NAME: &str = env!("CARGO_PKG_NAME");
 
@@ -26,12 +28,14 @@ impl Loader {
 
     pub fn load_path(&self, path: &Path) -> Result<Config> {
         let data = std::fs::read_to_string(&path).context("Config file")?;
-        toml::from_str(&data).context("Config Toml")
+        Ok(toml::from_str(&data).context("Config Toml")?)
     }
 
-    pub fn save_path(&self, path: &Path, cfg: &Config) -> Result<()> {
+    pub fn print(&self, cfg: &Config) -> Result<()> {
         let toml = toml::to_string_pretty(cfg).context("toml serialize")?;
-        std::fs::write(path, toml).context("write config")?;
+        std::io::stdout()
+            .write_all(toml.as_bytes())
+            .expect("Unable to write data");
         Ok(())
     }
 }
@@ -47,7 +51,21 @@ where
     })
 }
 
-#[derive(Debug, Clone)]
+pub fn option_explicit_serialize<T, S>(
+    val: &Option<T>,
+    serializer: S,
+) -> std::result::Result<S::Ok, S::Error>
+where
+    S: Serializer,
+    T: Serialize,
+{
+    match val {
+        None => str::serialize("none", serializer),
+        Some(ref val) => T::serialize(val, serializer),
+    }
+}
+
+#[derive(Debug, Clone, Copy)]
 pub struct Rgba {
     pub red: u8,
     pub green: u8,
@@ -74,14 +92,14 @@ impl<'de> Deserialize<'de> for Rgba {
 }
 
 impl std::str::FromStr for Rgba {
-    type Err = anyhow::Error;
+    type Err = Error;
     fn from_str(s: &str) -> std::result::Result<Self, Self::Err> {
         log::trace!("rgba::from_str {}", s);
         let without_prefix = s.trim_start_matches('#');
         match without_prefix.len() {
             8 => {
                 let mut bytes = [0u8; 4];
-                hex::decode_to_slice(without_prefix, &mut bytes)?;
+                hex::decode_to_slice(without_prefix, &mut bytes).context("color")?;
                 log::trace!("rgba::from_str {:?}", bytes);
                 Ok(Rgba {
                     red: bytes[0],
@@ -92,7 +110,7 @@ impl std::str::FromStr for Rgba {
             }
             6 => {
                 let mut bytes = [0u8; 3];
-                hex::decode_to_slice(without_prefix, &mut bytes)?;
+                hex::decode_to_slice(without_prefix, &mut bytes).context("color")?;
                 log::trace!("rgba::from_str {:?}", bytes);
                 Ok(Rgba {
                     red: bytes[0],
@@ -111,6 +129,7 @@ impl std::str::FromStr for Rgba {
 pub struct Config {
     pub title: String,
     pub grab_keyboard: bool,
+    pub resizable: bool,
     pub depth: u8,
     pub dialog: Dialog,
 }
@@ -118,8 +137,9 @@ pub struct Config {
 impl Default for Config {
     fn default() -> Self {
         Self {
-            title: "Passphrase request".into(),
+            title: NAME.into(),
             grab_keyboard: false,
+            resizable: false,
             depth: 24,
             dialog: Dialog::default(),
         }
@@ -129,11 +149,18 @@ impl Default for Config {
 #[derive(Debug, Serialize, Deserialize)]
 #[serde(default)]
 pub struct Dialog {
+    #[serde(serialize_with = "option_explicit_serialize")]
     #[serde(deserialize_with = "option_explicit_none")]
-    pub dpi: Option<f64>,
-    pub font: String,
+    pub font: Option<String>,
+    #[serde(serialize_with = "option_explicit_serialize")]
+    #[serde(deserialize_with = "option_explicit_none")]
+    pub font_file: Option<std::ffi::CString>,
     pub label: String,
+    #[serde(serialize_with = "option_explicit_serialize")]
+    #[serde(deserialize_with = "option_explicit_none")]
+    pub scale: Option<f64>,
     pub indicator_label: String,
+    #[serde(serialize_with = "option_explicit_serialize")]
     #[serde(deserialize_with = "option_explicit_none")]
     pub input_timeout: Option<u64>,
     pub foreground: Rgba,
@@ -172,19 +199,16 @@ impl Default for Dialog {
             label: "Please enter your authentication passphrase:".into(),
             indicator_label: "Secret:".into(),
             input_timeout: Some(30),
-            dpi: None,
-            font: "sans serif 11".into(),
+            font: Some("11".into()),
+            scale: None,
+            font_file: None,
             layout_opts: Layout::default(),
             ok_button,
             cancel_button,
             plaintext_button,
             clipboard_button: ClipboardButton {
                 foreground: "#5c616c".parse().unwrap(),
-                button: Button {
-                    horizontal_spacing: 10.0,
-                    vertical_spacing: 6.0,
-                    ..button
-                },
+                button,
             },
             indicator: Indicator::default(),
         }
@@ -229,8 +253,12 @@ impl Default for TextButton {
 #[derive(Debug, Clone, Serialize, Deserialize)]
 #[serde(default)]
 pub struct Button {
-    pub horizontal_spacing: f64,
-    pub vertical_spacing: f64,
+    #[serde(deserialize_with = "option_explicit_none")]
+    #[serde(serialize_with = "option_explicit_serialize")]
+    pub horizontal_spacing: Option<f64>,
+    #[serde(serialize_with = "option_explicit_serialize")]
+    #[serde(deserialize_with = "option_explicit_none")]
+    pub vertical_spacing: Option<f64>,
     pub border_width: f64,
     pub radius_x: f64,
     pub radius_y: f64,
@@ -239,11 +267,14 @@ pub struct Button {
     pub background: Rgba,
     pub border_color: Rgba,
     pub border_color_pressed: Rgba,
+    #[serde(serialize_with = "option_explicit_serialize")]
     #[serde(deserialize_with = "option_explicit_none")]
     pub background_stop: Option<Rgba>,
     pub background_pressed: Rgba,
+    #[serde(serialize_with = "option_explicit_serialize")]
     #[serde(deserialize_with = "option_explicit_none")]
     pub background_pressed_stop: Option<Rgba>,
+    #[serde(serialize_with = "option_explicit_serialize")]
     #[serde(deserialize_with = "option_explicit_none")]
     pub background_hover_stop: Option<Rgba>,
     pub background_hover: Rgba,
@@ -258,8 +289,8 @@ impl Default for Button {
             background_pressed_stop: None,
             background_hover: "#ffffff".parse().unwrap(),
             background_hover_stop: None,
-            horizontal_spacing: 16.0,
-            vertical_spacing: 6.0,
+            horizontal_spacing: None,
+            vertical_spacing: None,
             border_width: 1.0,
             border_color: "#cfd6e6".parse().unwrap(),
             border_color_pressed: "#b7c0d3".parse().unwrap(),
@@ -275,18 +306,40 @@ impl Default for Button {
 #[serde(default)]
 pub struct Layout {
     pub layout: crate::dialog::layout::Layout,
-    pub horizontal_spacing: f64,
-    pub vertical_spacing: f64,
+    #[serde(serialize_with = "option_explicit_serialize")]
+    #[serde(deserialize_with = "option_explicit_none")]
+    horizontal_spacing: Option<f64>,
+    #[serde(serialize_with = "option_explicit_serialize")]
+    #[serde(deserialize_with = "option_explicit_none")]
+    vertical_spacing: Option<f64>,
+    #[serde(serialize_with = "option_explicit_serialize")]
     #[serde(deserialize_with = "option_explicit_none")]
     pub text_width: Option<u32>,
 }
 
+impl Layout {
+    pub fn horizontal_spacing(&self, text_height: u32) -> f64 {
+        if let Some(horizontal_spacing) = self.horizontal_spacing {
+            horizontal_spacing
+        } else {
+            (text_height as f64 / 1.7).round()
+        }
+    }
+    pub fn vertical_spacing(&self, text_height: u32) -> f64 {
+        if let Some(vertical_spacing) = self.vertical_spacing {
+            vertical_spacing
+        } else {
+            (text_height as f64 / 1.7).round()
+        }
+    }
+}
+
 impl Default for Layout {
     fn default() -> Self {
         Self {
             layout: crate::dialog::layout::Layout::Center,
-            horizontal_spacing: 10.0,
-            vertical_spacing: 10.0,
+            horizontal_spacing: None,
+            vertical_spacing: None,
             text_width: None,
         }
     }
@@ -300,9 +353,12 @@ pub struct IndicatorClassic {
     pub radius_x: f64,
     pub radius_y: f64,
     #[serde(deserialize_with = "option_explicit_none")]
+    #[serde(serialize_with = "option_explicit_serialize")]
     pub horizontal_spacing: Option<f64>,
+    #[serde(serialize_with = "option_explicit_serialize")]
     #[serde(deserialize_with = "option_explicit_none")]
     pub element_height: Option<f64>,
+    #[serde(serialize_with = "option_explicit_serialize")]
     #[serde(deserialize_with = "option_explicit_none")]
     pub element_width: Option<f64>,
 }
@@ -324,6 +380,7 @@ impl Default for IndicatorClassic {
 #[derive(Debug, Clone, Serialize, Deserialize)]
 #[serde(default)]
 pub struct IndicatorCircle {
+    #[serde(serialize_with = "option_explicit_serialize")]
     #[serde(deserialize_with = "option_explicit_none")]
     pub diameter: Option<f64>,
     pub rotate: bool,
@@ -332,6 +389,7 @@ pub struct IndicatorCircle {
     pub light_up: bool,
     pub spacing_angle: f64,
     pub indicator_count: u32,
+    #[serde(serialize_with = "option_explicit_serialize")]
     #[serde(deserialize_with = "option_explicit_none")]
     pub indicator_width: Option<f64>,
     pub lock_color: Rgba,
@@ -343,8 +401,8 @@ impl Default for IndicatorCircle {
             diameter: None,
             rotate: true,
             light_up: true,
-            rotation_speed_start: 0.05,
-            rotation_speed_gain: 2.0,
+            rotation_speed_start: 0.10,
+            rotation_speed_gain: 1.2,
             spacing_angle: 0.5,
             indicator_count: 3,
             indicator_width: None,
@@ -450,8 +508,12 @@ pub enum StringType {
 #[derive(Debug, Clone, Serialize, Deserialize)]
 #[serde(default)]
 pub struct IndicatorStrings {
-    pub horizontal_spacing: f64,
-    pub vertical_spacing: f64,
+    #[serde(serialize_with = "option_explicit_serialize")]
+    #[serde(deserialize_with = "option_explicit_none")]
+    pub horizontal_spacing: Option<f64>,
+    #[serde(serialize_with = "option_explicit_serialize")]
+    #[serde(deserialize_with = "option_explicit_none")]
+    pub vertical_spacing: Option<f64>,
     pub radius_x: f64,
     pub radius_y: f64,
     #[serde(flatten)]
@@ -461,8 +523,8 @@ pub struct IndicatorStrings {
 impl Default for IndicatorStrings {
     fn default() -> Self {
         Self {
-            horizontal_spacing: 8.0,
-            vertical_spacing: 6.0,
+            horizontal_spacing: None,
+            vertical_spacing: None,
             radius_x: 2.0,
             radius_y: 2.0,
             strings: StringType::Asterisk {
@@ -502,8 +564,8 @@ pub struct Indicator {
 impl Default for Indicator {
     fn default() -> Self {
         Self {
-            indicator_type: IndicatorType::Strings {
-                strings: IndicatorStrings::default(),
+            indicator_type: IndicatorType::Circle {
+                circle: IndicatorCircle::default(),
             },
             common: IndicatorCommon::default(),
         }
@@ -518,11 +580,13 @@ pub struct IndicatorCommon {
     pub foreground: Rgba,
     pub background: Rgba,
     #[serde(deserialize_with = "option_explicit_none")]
+    #[serde(serialize_with = "option_explicit_serialize")]
     pub background_stop: Option<Rgba>,
     pub border_color: Rgba,
     pub border_color_focused: Rgba,
     pub indicator_color: Rgba,
     #[serde(deserialize_with = "option_explicit_none")]
+    #[serde(serialize_with = "option_explicit_serialize")]
     pub indicator_color_stop: Option<Rgba>,
 }
 
