@@ -4,7 +4,7 @@ use std::ops::{Deref, DerefMut};
 use std::pin::Pin;
 use std::time::Duration;
 
-use log::trace;
+use log::{debug, trace};
 use rand::seq::SliceRandom as _;
 use tokio::time::{sleep, Instant, Sleep};
 
@@ -27,8 +27,8 @@ pub struct Base {
     border_pattern: Pattern,
     border_pattern_focused: Pattern,
     indicator_pattern: Pattern,
-    pub(super) dirty: bool,
-    pub(super) dirty_blink: bool,
+    dirty: bool,
+    dirty_blink: bool,
     blink_do: bool,
     blink_enabled: bool,
     blink_on: bool,
@@ -74,28 +74,28 @@ impl Base {
         }
     }
 
-    pub fn pass_delete(&mut self) -> bool {
-        let dirty = self.key_pressed();
+    pub fn dirty(&self) -> bool {
+        self.dirty || self.dirty_blink
+    }
+
+    pub fn pass_delete(&mut self) {
+        self.key_pressed();
         if self.pass.len > 0 {
             self.pass.len -= 1;
             self.dirty = true;
-            return true;
         }
-        dirty
     }
 
-    pub fn pass_clear(&mut self) -> bool {
-        let dirty = self.key_pressed();
+    pub fn pass_clear(&mut self) {
+        self.key_pressed();
         if self.pass.len != 0 {
             self.pass.len = 0;
             self.dirty = true;
-            return true;
         }
-        dirty
     }
 
-    pub fn pass_insert(&mut self, s: &str, pasted: bool) -> bool {
-        let dirty = self.key_pressed();
+    pub fn pass_insert(&mut self, s: &str, pasted: bool) {
+        self.key_pressed();
         let mut inserted = false;
         for c in s.chars() {
             if !self.pass.push(c) {
@@ -108,10 +108,8 @@ impl Base {
                 self.show_selection();
             }
             self.dirty = true;
-            return true;
         }
         trace!("pass insert failed");
-        dirty
     }
 
     pub fn init_timeouts(&mut self) {
@@ -125,7 +123,7 @@ impl Base {
         self.blink_do || self.show_selection_do
     }
 
-    pub async fn handle_events(&mut self) -> bool {
+    pub async fn handle_events(&mut self) {
         tokio::select! {
             _ = &mut self.blink_timeout, if self.blink_do => {
                 self.on_blink_timeout()
@@ -185,14 +183,13 @@ impl Base {
         cr.restore().unwrap();
     }
 
-    pub fn on_show_selection_timeout(&mut self) -> bool {
+    pub fn on_show_selection_timeout(&mut self) {
         assert!(self.show_selection_do);
         self.show_selection_do = false;
         self.dirty = true;
-        true
     }
 
-    fn show_selection(&mut self) -> bool {
+    fn show_selection(&mut self) {
         self.show_selection_do = true;
         self.show_selection_timeout.as_mut().reset(
             Instant::now()
@@ -200,23 +197,19 @@ impl Base {
                 .unwrap(),
         );
         self.dirty = true;
-        true
     }
 
-    pub fn key_pressed(&mut self) -> bool {
-        let mut ret = false;
+    pub fn key_pressed(&mut self) {
         if self.blink_enabled {
             if !self.blink_on {
                 self.dirty_blink = true;
-                ret = true;
             }
             self.blink_on = true;
             self.reset_blink();
         }
-        ret
     }
 
-    pub fn set_focused(&mut self, is_focused: bool) -> bool {
+    pub fn set_focused(&mut self, is_focused: bool) {
         self.dirty = self.dirty || is_focused != self.has_focus;
         self.has_focus = is_focused;
         if self.blink_enabled {
@@ -226,15 +219,13 @@ impl Base {
             }
             self.blink_do = is_focused;
         }
-        self.dirty
     }
 
-    pub fn on_blink_timeout(&mut self) -> bool {
+    pub fn on_blink_timeout(&mut self) {
         trace!("blink timeout");
         self.blink_on = !self.blink_on;
         self.dirty_blink = true;
         self.reset_blink();
-        true
     }
 
     fn reset_blink(&mut self) {
@@ -326,28 +317,25 @@ impl Circle {
         }
     }
 
-    pub fn pass_delete(&mut self) -> bool {
-        let ret = self.base.pass_delete();
+    pub fn pass_delete(&mut self) {
+        self.base.pass_delete();
         if self.rotate {
             self.init_rotation();
         }
-        ret
     }
 
-    pub fn pass_clear(&mut self) -> bool {
-        let ret = self.base.pass_clear();
+    pub fn pass_clear(&mut self) {
+        self.base.pass_clear();
         if self.rotate {
             self.init_rotation();
         }
-        ret
     }
 
-    pub fn pass_insert(&mut self, s: &str, pasted: bool) -> bool {
-        let ret = self.base.pass_insert(s, pasted);
+    pub fn pass_insert(&mut self, s: &str, pasted: bool) {
+        self.base.pass_insert(s, pasted);
         if self.rotate {
             self.init_rotation();
         }
-        ret
     }
 
     pub fn into_pass(self) -> Passphrase {
@@ -439,7 +427,7 @@ impl Circle {
     }
 
     // TODO
-    pub fn update(&self, cr: &cairo::Context, background: &super::Pattern) {
+    pub fn repaint(&self, cr: &cairo::Context, background: &super::Pattern) {
         if self.dirty {
             trace!("indicator dirty");
             self.clear(cr, background);
@@ -653,7 +641,7 @@ impl Classic {
     }
 
     // TODO
-    pub fn update(&self, cr: &cairo::Context, background: &super::Pattern) {
+    pub fn repaint(&self, cr: &cairo::Context, background: &super::Pattern) {
         if self.dirty {
             trace!("indicator dirty");
             self.clear(cr, background);
@@ -732,14 +720,6 @@ impl StringType {
             Self::Asterisk(asterisk) => asterisk.set_text(layout, pass),
         }
     }
-
-    pub fn height(&self) -> i32 {
-        match self {
-            Self::Disco(disco) => disco.height,
-            Self::Custom(custom) => custom.height,
-            Self::Asterisk(asterisk) => asterisk.height,
-        }
-    }
 }
 
 #[derive(Debug)]
@@ -781,6 +761,7 @@ impl Strings {
         strings_cfg: config::IndicatorStrings,
         layout: pango::Layout,
         debug: bool,
+        text_height: u32,
     ) -> Result<Self> {
         let strings = match strings_cfg.strings {
             config::StringType::Asterisk { asterisk } => {
@@ -791,14 +772,24 @@ impl Strings {
                 StringType::Custom(Custom::new(custom, &layout))
             }
         };
-        let height = strings.height() as f64
-            + 2.0 * strings_cfg.vertical_spacing
-            + 2.0 * config.border_width;
+        let vertical_spacing = strings_cfg
+            .vertical_spacing
+            .unwrap_or(text_height as f64 / 3.0)
+            .round();
+        let horizontal_spacing = strings_cfg
+            .horizontal_spacing
+            .unwrap_or(text_height as f64 / 2.0)
+            .round();
+        debug!(
+            "strings indicator: vertical_spacing: {}, horizontal_spacing: {}, border_width: {}",
+            vertical_spacing, horizontal_spacing, config.border_width
+        );
+        let height = text_height as f64 + 2.0 * vertical_spacing + 2.0 * config.border_width;
         let base = Base {
             ..Base::new(config, height, debug)
         };
 
-        layout.set_height(strings.height() * pango::SCALE);
+        layout.set_height(i32::try_from(text_height).unwrap() * pango::SCALE);
         layout.set_single_paragraph_mode(true);
 
         let blink_spacing = if strings.use_cursor() { 0.0 } else { 8.0 };
@@ -807,8 +798,8 @@ impl Strings {
             strings,
             radius_x: strings_cfg.radius_x,
             radius_y: strings_cfg.radius_x,
-            horizontal_spacing: strings_cfg.horizontal_spacing,
-            vertical_spacing: strings_cfg.vertical_spacing,
+            horizontal_spacing,
+            vertical_spacing,
             blink_spacing,
             index: (0, 0),
             layout,
@@ -837,21 +828,19 @@ impl Strings {
         Ok(())
     }
 
-    pub fn pass_clear(&mut self) -> bool {
-        let dirty = self.key_pressed();
+    pub fn pass_clear(&mut self) {
+        self.key_pressed();
         if self.pass.len != 0 {
             self.cursor = 0;
             self.pass.len = 0;
             self.set_text();
             self.dirty = true;
-            return true;
         }
-        dirty
     }
 
-    pub fn pass_insert(&mut self, s: &str, pasted: bool) -> bool {
+    pub fn pass_insert(&mut self, s: &str, pasted: bool) {
         trace!("pass insert {}", self.cursor);
-        let ret = self.base.key_pressed();
+        self.base.key_pressed();
         let cursor = self.cursor;
         let inserted = self.pass.insert_many(cursor, s.chars(), s.chars().count());
         if inserted > 0 {
@@ -862,31 +851,28 @@ impl Strings {
             self.cursor += inserted;
             self.dirty = true;
             trace!("pass inserted");
-            return true;
         }
-        ret
     }
 
-    pub fn pass_delete(&mut self) -> bool {
+    pub fn pass_delete(&mut self) {
         trace!("pass delete {}", self.cursor);
-        let ret = self.base.key_pressed();
+        self.base.key_pressed();
         if self.cursor == 0 {
-            return ret;
+            return;
         }
         let i = self.cursor - 1;
         self.pass.delete(i);
         self.set_text();
         self.cursor -= 1;
         self.dirty = true;
-        true
     }
 
-    pub fn move_cursor(&mut self, direction: i32) -> bool {
+    pub fn move_cursor(&mut self, direction: i32) {
         if !self.strings.use_cursor() && !self.show_plain {
-            return false;
+            return;
         }
         trace!("move cursor {}", self.cursor);
-        let dirty = self.key_pressed();
+        self.key_pressed();
         let new_cursor = self
             .layout
             .move_cursor_visually(true, self.cursor_bytes(), 0, direction);
@@ -894,9 +880,7 @@ impl Strings {
             self.cursor = self.cursor_chars(new_cursor.0, new_cursor.1);
 
             self.dirty = true;
-            return true;
         }
-        dirty
     }
 
     fn cursor_chars(&self, idx: i32, trailing: i32) -> usize {
@@ -944,7 +928,7 @@ impl Strings {
     }
 
     pub fn paint(&self, cr: &cairo::Context) {
-        trace!("paint start");
+        trace!("strings paint start");
         assert!(self.width != 0.0);
         cr.save().unwrap();
         cr.translate(self.x, self.y);
@@ -989,13 +973,12 @@ impl Strings {
         trace!("paint end");
     }
 
-    pub fn on_show_selection_timeout(&mut self) -> bool {
+    pub fn on_show_selection_timeout(&mut self) {
         self.base.on_show_selection_timeout();
         self.set_text();
-        true
     }
 
-    pub async fn handle_events(&mut self) -> bool {
+    pub async fn handle_events(&mut self) {
         tokio::select! {
             _ = &mut self.base.blink_timeout, if self.base.blink_do => {
                 self.on_blink_timeout()
@@ -1007,7 +990,7 @@ impl Strings {
     }
 
     // TODO
-    pub fn update(&self, cr: &cairo::Context, background: &super::Pattern) {
+    pub fn repaint(&self, cr: &cairo::Context, background: &super::Pattern) {
         if self.dirty {
             trace!("indicator dirty");
             self.clear(cr, background);
@@ -1018,10 +1001,10 @@ impl Strings {
         }
     }
 
-    // return (inside, dirty)
-    pub fn set_cursor(&mut self, x: f64, y: f64) -> (bool, bool) {
+    // return is_inside
+    pub fn set_cursor(&mut self, x: f64, y: f64) -> bool {
         if !self.show_plain && !self.strings.use_cursor() {
-            return (false, false);
+            return false;
         }
 
         if self.is_inside(x, y) {
@@ -1051,7 +1034,7 @@ impl Strings {
                 self.key_pressed();
                 self.cursor = self.cursor_chars(idx, trailing);
                 self.dirty = true;
-                return (true, true);
+                return true;
             } else {
                 assert!(
                     self.pass.len == 0,
@@ -1062,10 +1045,10 @@ impl Strings {
                     idx,
                     trailing
                 );
-                return (false, false);
+                return false;
             }
         }
-        (false, false)
+        false
     }
 
     fn set_text(&mut self) {
@@ -1110,24 +1093,21 @@ impl Strings {
 
 #[derive(Debug)]
 struct Custom {
-    height: i32,
     width: i32,
     strings: Vec<String>,
 }
 
 impl Custom {
     pub fn new(config: config::Custom, layout: &pango::Layout) -> Self {
-        let sizes: Vec<(i32, i32)> = config
+        let width = config
             .strings
             .iter()
             .map(|s| {
                 layout.set_text(s);
-                layout.pixel_size()
+                layout.pixel_size().0
             })
-            .collect();
-        // every string with the same font should have the same logical height
-        let height = sizes[0].1;
-        let width = sizes.into_iter().map(|(w, _)| w).max().unwrap();
+            .max()
+            .unwrap();
         layout.set_width(width * pango::SCALE);
         layout.set_alignment(config.alignment.into());
         layout.set_justify(config.justify);
@@ -1136,11 +1116,7 @@ impl Custom {
             let mut rand = rand::thread_rng();
             strings[1..].shuffle(&mut rand);
         }
-        Self {
-            height,
-            width,
-            strings,
-        }
+        Self { width, strings }
     }
 
     pub fn set_text(&mut self, layout: &pango::Layout, pass: &SecBuf<char>, show_paste: bool) {
@@ -1160,7 +1136,6 @@ impl Custom {
 
 #[derive(Debug)]
 struct Disco {
-    height: i32,
     widths: Vec<f64>,
     dancer_max_width: f64,
     separator_width: f64,
@@ -1183,13 +1158,11 @@ impl Disco {
             })
             .collect();
         // every string with the same font should have the same logical height
-        let height = sizes[0].1;
         let widths = sizes.iter().map(|(w, _)| *w as f64).collect();
         let dancer_max_width = sizes.into_iter().map(|(w, _)| w).max().unwrap() as f64;
         layout.set_text("");
         trace!("disco new end");
         Self {
-            height,
             widths,
             dancer_max_width,
             separator_width: layout.pixel_size().1 as f64,
@@ -1257,7 +1230,6 @@ impl Disco {
 
 #[derive(Debug)]
 struct Asterisk {
-    height: i32,
     asterisk_width: f64,
     asterisk: String,
     count: u16,
@@ -1269,11 +1241,10 @@ impl Asterisk {
     pub fn new(config: config::Asterisk, layout: &pango::Layout) -> Self {
         let asterisk: String = config.asterisk;
         layout.set_text(&asterisk);
-        let (asterisk_width, height) = layout.pixel_size();
+        let (asterisk_width, _) = layout.pixel_size();
         layout.set_alignment(config.alignment.into());
         layout.set_text("");
         Self {
-            height,
             asterisk_width: asterisk_width as f64,
             asterisk,
             min_count: config.min_count,
