@@ -2,21 +2,23 @@ use std::convert::TryFrom as _;
 use std::convert::TryInto as _;
 use std::ffi::CStr;
 use std::ops::{Deref, DerefMut};
+use std::pin::Pin;
 use std::time::Duration;
 
+use fontconfig_sys::fontconfig;
 use libc::LC_ALL;
 use log::{debug, info, log_enabled, trace, warn};
 use pango::prelude::FontExt as _;
-use tokio::time::{sleep, Instant};
-use x11rb::connection::Connection as _;
+use tokio::time::{sleep, Instant, Sleep};
 use x11rb::protocol::xproto;
 use zeroize::Zeroize;
 
 use crate::backbuffer::FrameId;
+use crate::bail;
 use crate::config;
 use crate::config::{IndicatorType, Rgba};
 use crate::errors::Result;
-use crate::event::{Event, Keypress, XContext};
+use crate::event::{Keypress, XContext};
 use crate::keyboard::{
     self, keysyms, xkb_compose_feed_result, xkb_compose_status, Keyboard, Keycode,
 };
@@ -75,8 +77,11 @@ impl Components {
                 config.foreground.into(),
                 self.text_height as f64,
             ));
-            self.buttons
-                .push(Button::new(config.button, clipboard_label));
+            self.buttons.push(Button::new(
+                config.button,
+                clipboard_label,
+                self.text_height,
+            ));
         }
         &mut self.buttons[2]
     }
@@ -88,7 +93,8 @@ impl Components {
             let layout = pango::Layout::new(&self.pango_context);
             layout.set_text(&config.label);
             let label = Label::TextLabel(TextLabel::new(config.foreground.into(), layout));
-            self.buttons.push(Button::new(config.button, label));
+            self.buttons
+                .push(Button::new(config.button, label, self.text_height));
         }
         &mut self.buttons[3]
     }
@@ -117,7 +123,7 @@ pub enum Pattern {
 }
 
 impl Pattern {
-    fn get_pattern(fill_height: f64, start: Rgba, end: Option<Rgba>) -> Self {
+    pub fn get_pattern(fill_height: f64, start: Rgba, end: Option<Rgba>) -> Self {
         if let Some(end) = end {
             let grad = cairo::LinearGradient::new(0.0, 0.0, 0.0, fill_height);
             grad.add_color_stop_rgba(
@@ -187,7 +193,7 @@ impl Indicator {
         }
     }
 
-    pub async fn handle_events(&mut self) -> bool {
+    pub async fn handle_events(&mut self) {
         match self {
             Self::Strings(i) => i.handle_events().await,
             Self::Circle(i) => i.handle_events().await,
@@ -195,7 +201,7 @@ impl Indicator {
         }
     }
 
-    pub fn pass_insert(&mut self, s: &str, pasted: bool) -> bool {
+    pub fn pass_insert(&mut self, s: &str, pasted: bool) {
         match self {
             Self::Strings(i) => i.pass_insert(s, pasted),
             Self::Circle(i) => i.pass_insert(s, pasted),
@@ -203,7 +209,7 @@ impl Indicator {
         }
     }
 
-    pub fn pass_clear(&mut self) -> bool {
+    pub fn pass_clear(&mut self) {
         match self {
             Self::Strings(i) => i.pass_clear(),
             Self::Circle(i) => i.pass_clear(),
@@ -211,7 +217,7 @@ impl Indicator {
         }
     }
 
-    pub fn pass_delete(&mut self) -> bool {
+    pub fn pass_delete(&mut self) {
         match self {
             Self::Strings(i) => i.pass_delete(),
             Self::Circle(i) => i.pass_delete(),
@@ -219,19 +225,19 @@ impl Indicator {
         }
     }
 
-    pub fn move_cursor(&mut self, direction: i32) -> bool {
+    pub fn move_cursor(&mut self, direction: i32) {
         match self {
             Self::Strings(i) => i.move_cursor(direction),
-            Self::Circle(..) => false,
-            Self::Classic(..) => false,
+            Self::Circle(..) => {}
+            Self::Classic(..) => {}
         }
     }
 
-    pub fn set_cursor(&mut self, x: f64, y: f64) -> (bool, bool) {
+    pub fn set_cursor(&mut self, x: f64, y: f64) -> bool {
         match self {
             Self::Strings(i) => i.set_cursor(x, y),
-            Self::Circle(..) => (false, false),
-            Self::Classic(..) => (false, false),
+            Self::Circle(..) => false,
+            Self::Classic(..) => false,
         }
     }
 
@@ -285,11 +291,11 @@ impl Indicator {
         }
     }
 
-    pub fn update(&self, cr: &cairo::Context, bg: &Pattern) {
+    pub fn repaint(&self, cr: &cairo::Context, bg: &Pattern) {
         match self {
-            Self::Strings(i) => i.update(cr, bg),
-            Self::Circle(i) => i.update(cr, bg),
-            Self::Classic(i) => i.update(cr, bg),
+            Self::Strings(i) => i.repaint(cr, bg),
+            Self::Circle(i) => i.repaint(cr, bg),
+            Self::Classic(i) => i.repaint(cr, bg),
         }
     }
 
@@ -392,7 +398,7 @@ impl ClipboardLabel {
                 x: 0.0,
                 y: 0.0,
                 height: text_height,
-                width: (text_height * 0.85).round(),
+                width: text_height * 0.83,
             },
             foreground,
         }
@@ -401,12 +407,10 @@ impl ClipboardLabel {
         cr.save().unwrap();
         cr.translate(self.rectangle.x, self.rectangle.y);
 
-        let line_width = 1.5;
-        let barely_noticeable = (self.rectangle.width / 10.0).floor().max(1.0);
-        let small_height = ((self.rectangle.width - 4.0 * barely_noticeable - 2.0 * line_width)
-            * 0.8)
-            .round()
-            .max(2.0);
+        let dot = self.rectangle.height / 18.0;
+        let line_width = dot * 1.5;
+        let small_height =
+            ((self.rectangle.width - 4.0 * dot - 2.0 * line_width) * 0.8).max(2.0 * dot);
         cr.rectangle(0.0, 0.0, self.rectangle.width, self.rectangle.height);
         cr.rectangle(
             line_width,
@@ -417,11 +421,11 @@ impl ClipboardLabel {
         cr.set_fill_rule(cairo::FillRule::EvenOdd);
         cr.clip();
 
-        let y_offset = barely_noticeable;
+        let y_offset = dot;
         Button::rounded_rectangle(
             cr,
-            2.0,
-            2.0,
+            2.0 * dot,
+            2.0 * dot,
             line_width / 2.0,
             line_width / 2.0 + y_offset,
             self.rectangle.width - line_width,
@@ -432,9 +436,9 @@ impl ClipboardLabel {
         cr.stroke().unwrap();
 
         cr.reset_clip();
-        let small_width = self.rectangle.width - 4.0 * barely_noticeable - 3.0 * line_width;
+        let small_width = self.rectangle.width - 4.0 * dot - 3.0 * line_width;
         cr.rectangle(
-            line_width + barely_noticeable * 2.0 + line_width / 2.0,
+            line_width + dot * 2.0 + line_width / 2.0,
             line_width / 2.0,
             small_width,
             small_height - line_width,
@@ -476,7 +480,7 @@ impl TextLabel {
         } else {
             self.layout.pixel_extents().1
         };
-        trace!("label rect: {:?}", rect);
+        debug!("label rect: {:?}", rect);
         let mut width: u32 = rect.width.try_into().unwrap();
         let mut height: u32 = rect.height.try_into().unwrap();
 
@@ -544,17 +548,12 @@ pub struct Button {
     pressed: bool,
     hover: bool,
     dirty: bool,
-    horizontal_spacing: f64,
-    vertical_spacing: f64,
-    border_width: f64,
     border_pattern: Pattern,
     border_pattern_pressed: Pattern,
+    vertical_spacing: f64,
+    horizontal_spacing: f64,
     interior_width: f64,
     interior_height: f64,
-    pressed_adjustment_x: f64,
-    pressed_adjustment_y: f64,
-    radius_x: f64,
-    radius_y: f64,
     label: Label,
     background: Option<Pattern>,
     bg_pressed: Option<Pattern>,
@@ -564,7 +563,26 @@ pub struct Button {
 }
 
 impl Button {
-    pub fn new(config: config::Button, label: Label) -> Self {
+    pub fn new(config: config::Button, label: Label, text_height: u32) -> Self {
+        let vertical_spacing = config
+            .vertical_spacing
+            .unwrap_or(text_height as f64 / 3.0)
+            .round();
+        let horizontal_spacing = if matches!(label, Label::ClipboardLabel(_)) {
+            config
+                .horizontal_spacing
+                .unwrap_or(text_height as f64 / 2.0)
+                .round()
+        } else {
+            config
+                .horizontal_spacing
+                .unwrap_or(text_height as f64)
+                .round()
+        };
+        debug!(
+            "button vertical_spacing: {}, horizontal_spacing: {}, border_width: {}",
+            vertical_spacing, horizontal_spacing, config.border_width
+        );
         let mut me = Self {
             x: 0.0,
             y: 0.0,
@@ -573,17 +591,12 @@ impl Button {
             pressed: false,
             hover: false,
             dirty: true,
-            horizontal_spacing: config.horizontal_spacing,
-            vertical_spacing: config.vertical_spacing,
-            border_width: config.border_width,
-            border_pattern: config.border_color.clone().into(), // TODO avoid cloning?
-            border_pattern_pressed: config.border_color_pressed.clone().into(), // TODO avoid cloning?
-            radius_x: config.radius_x,
-            radius_y: config.radius_y,
+            border_pattern: config.border_color.into(),
+            border_pattern_pressed: config.border_color_pressed.into(),
             interior_width: 0.0,
             interior_height: 0.0,
-            pressed_adjustment_x: config.pressed_adjustment_x,
-            pressed_adjustment_y: config.pressed_adjustment_y,
+            vertical_spacing,
+            horizontal_spacing,
             label,
             background: None,
             bg_pressed: None,
@@ -619,38 +632,42 @@ impl Button {
     }
 
     fn calc_total_extents(&mut self) {
-        self.width = self.interior_width + 2.0 * self.border_width;
-        self.height = self.interior_height + 2.0 * self.border_width;
+        self.width = self.interior_width + 2.0 * self.config.border_width;
+        self.height = self.interior_height + 2.0 * self.config.border_width;
 
-        // TODO placement, avoid cloning
-        let fill_height = self.height - self.border_width;
+        // TODO placement
+        let fill_height = self.height - self.config.border_width;
         self.background = Some(Pattern::get_pattern(
             fill_height,
-            self.config.background.clone(),
-            self.config.background_stop.clone(),
+            self.config.background,
+            self.config.background_stop,
         ));
         self.bg_pressed = Some(Pattern::get_pattern(
             fill_height,
-            self.config.background_pressed.clone(),
-            self.config.background_pressed_stop.clone(),
+            self.config.background_pressed,
+            self.config.background_pressed_stop,
         ));
         self.bg_hover = Some(Pattern::get_pattern(
             fill_height,
-            self.config.background_hover.clone(),
-            self.config.background_hover_stop.clone(),
+            self.config.background_hover,
+            self.config.background_hover_stop,
         ));
     }
 
     fn calc_label_position(&mut self) {
-        self.label.x = ((self.width - self.label.width) / 2.0).floor();
-        self.label.y = ((self.height - self.label.height) / 2.0).floor();
+        self.label.x = (self.width - self.label.width) / 2.0;
+        self.label.y = (self.height - self.label.height) / 2.0;
+        debug!(
+            "button/label: label.x: {}, label.y: {}",
+            self.label.x, self.label.y
+        );
     }
 
     pub fn is_inside(&self, x: f64, y: f64) -> bool {
-        x >= self.x + self.border_width
-            && x < self.x + self.width - self.border_width
-            && y >= self.y + self.border_width
-            && y < self.y + self.height - self.border_width
+        x >= self.x + self.config.border_width
+            && x < self.x + self.width - self.config.border_width
+            && y >= self.y + self.config.border_width
+            && y < self.y + self.height - self.config.border_width
     }
 
     pub fn set_hover(&mut self, hover: bool) {
@@ -673,15 +690,16 @@ impl Button {
         w: f64,
         h: f64,
     ) {
+        trace!("rounded_rectangle x: {}, y: {}, w: {}, h: {}", x, y, w, h);
         // from mono moonlight aka mono silverlight
         // test limits (without using multiplications)
         // http://graphics.stanford.edu/courses/cs248-98-fall/Final/q1.html
         const ARC_TO_BEZIER: f64 = 0.55228475;
         if radius_x > w - radius_x {
-            radius_x = (w / 2.0).floor();
+            radius_x = w / 2.0;
         }
         if radius_y > h - radius_y {
-            radius_y = (h / 2.0).floor();
+            radius_y = h / 2.0;
         }
 
         // approximate (quite close) the arc using a bezier curve
@@ -701,18 +719,31 @@ impl Button {
         cr.close_path();
     }
 
-    pub fn paint(&mut self, cr: &cairo::Context) {
+    pub fn set_painted(&mut self) {
+        self.dirty = false;
+    }
+
+    pub fn paint(&self, cr: &cairo::Context) {
+        trace!("button paint start");
         cr.save().unwrap();
         cr.translate(self.x, self.y);
 
         // "Note that while stroking the path transfers the source for half of the line width on
         // each side of the path, filling a path fills directly up to the edge of the path and no
         // further." We use stroke below so modifying accordingly.
-        let x = self.border_width / 2.0;
-        let y = self.border_width / 2.0;
-        let width = self.width - self.border_width;
-        let height = self.height - self.border_width;
-        Self::rounded_rectangle(cr, self.radius_x, self.radius_y, x, y, width, height);
+        let x = self.config.border_width / 2.0;
+        let y = self.config.border_width / 2.0;
+        let width = self.width - self.config.border_width;
+        let height = self.height - self.config.border_width;
+        Self::rounded_rectangle(
+            cr,
+            self.config.radius_x,
+            self.config.radius_y,
+            x,
+            y,
+            width,
+            height,
+        );
 
         let bg = if self.pressed && self.hover {
             &self.bg_pressed
@@ -726,23 +757,25 @@ impl Button {
         cr.set_source(bg.as_ref().unwrap()).unwrap();
         cr.fill_preserve().unwrap();
 
-        if self.border_width > 0.0 {
+        if self.config.border_width > 0.0 {
             if std::ptr::eq(bg, &self.bg_pressed) {
                 cr.set_source(&self.border_pattern_pressed).unwrap();
             } else {
                 cr.set_source(&self.border_pattern).unwrap();
             }
-            cr.set_line_width(self.border_width);
+            cr.set_line_width(self.config.border_width);
             cr.stroke().unwrap();
         }
 
         if self.pressed && self.hover {
-            cr.translate(self.pressed_adjustment_x, self.pressed_adjustment_y);
+            cr.translate(
+                self.config.pressed_adjustment_x,
+                self.config.pressed_adjustment_y,
+            );
         }
         self.label.paint(cr);
 
         cr.restore().unwrap();
-        self.dirty = false;
     }
 }
 
@@ -776,8 +809,10 @@ pub struct Dialog {
     height: f64,
     mouse_middle_pressed: bool,
     input_timeout_duration: Option<Duration>,
+    input_timeout: Option<Pin<Box<Sleep>>>,
     debug: bool,
-    pub cursor_size: Option<(u32, u32)>,
+    pub uses_cursor: bool,
+    button_pressed: bool,
 }
 
 impl Dialog {
@@ -788,31 +823,49 @@ impl Dialog {
         label: Option<&str>,
         debug: bool,
     ) -> Result<Self> {
-        let pango_context = pangocairo::create_context(cr).unwrap();
+        if let Some(font_file) = config.font_file {
+            unsafe {
+                let fc = fontconfig::FcConfigCreate();
+                if fontconfig::FcConfigSetCurrent(fc) == 0 {
+                    bail!("FcConfigSetCurrent failed");
+                }
+                if fontconfig::FcConfigAppFontAddFile(std::ptr::null_mut(), font_file.as_ptr() as _)
+                    == 0
+                {
+                    bail!("Could not load font file: {}", font_file.to_string_lossy());
+                }
+            }
+        }
 
-        let dpi = if let Some(dpi) = config.dpi {
-            dpi
-        } else {
-            (screen.height_in_pixels as f64 * 25.4 / screen.height_in_millimeters as f64)
-                .max(96.0)
-                .round()
-        };
-        debug!("dpi {}", dpi);
-        pangocairo::context_set_resolution(&pango_context, dpi);
+        if let Some(scale) = config.scale {
+            if scale <= 0.0 {
+                bail!("invalid scale {}", scale);
+            }
+            cr.scale(scale, scale);
+        } else if screen.height_in_pixels > 1080 {
+            let scale = screen.height_in_pixels as f64 / 1080.0;
+            cr.scale(scale, scale);
+        }
+
+        let pango_context = pangocairo::create_context(cr).unwrap();
 
         let language = pango::Language::default();
         debug!("language {}", language.to_string());
         pango_context.set_language(&language);
 
-        let font = config.font;
-        let font_desc = pango::FontDescription::from_string(&font);
+        if let Some(font) = config.font {
+            let mut font_desc = pango::FontDescription::from_string(&font);
+            debug!("font request: {}", font_desc.to_string());
+            if font_desc.size() == 0 {
+                debug!("setting font size to default 11");
+                font_desc.set_size(11 * pango::SCALE);
+            }
+            pango_context.set_font_description(&font_desc);
+        }
 
-        pango_context.set_font_description(&font_desc);
-
-        debug!("font request: {}", font_desc.to_string());
         if log_enabled!(log::Level::Debug) {
             let closest_font = pango_context
-                .load_font(&font_desc)
+                .load_font(&pango_context.font_description().unwrap())
                 .unwrap()
                 .describe()
                 .map(|f| f.to_string())
@@ -820,11 +873,14 @@ impl Dialog {
             debug!("closest font: {}", closest_font);
         }
 
-        let metrics = pango_context.metrics(Some(&font_desc), None).unwrap();
-        let text_height: u32 = ((metrics.ascent() + metrics.descent() + pango::SCALE - 1)
-            / pango::SCALE)
-            .try_into()
-            .unwrap();
+        let metrics = pango_context.metrics(None, None).unwrap();
+        let text_height = (metrics.ascent() + metrics.descent()) as f64 / pango::SCALE as f64;
+        let text_height = cr
+            .user_to_device_distance(0.0, text_height)
+            .expect("cairo user_to_device_distance")
+            .1
+            .ceil() as u32;
+        debug!("text height: {}", text_height);
 
         let label_layout = pango::Layout::new(&pango_context);
         label_layout.set_text(label.unwrap_or(&config.label));
@@ -845,28 +901,25 @@ impl Dialog {
             cancel_layout,
         ));
 
-        let mut ok_button = Button::new(config.ok_button.button, ok_label);
-        let mut cancel_button = Button::new(config.cancel_button.button, cancel_label);
+        let mut ok_button = Button::new(config.ok_button.button, ok_label, text_height);
+        let mut cancel_button = Button::new(config.cancel_button.button, cancel_label, text_height);
         balance_button_extents(&mut ok_button, &mut cancel_button);
 
         // TODO
-        let cursor_size = if matches!(
+        let uses_cursor = matches!(
             config.indicator.indicator_type,
             IndicatorType::Strings { .. }
-        ) {
-            Some((7, text_height))
-        } else {
-            None
-        };
+        );
 
         let mut indicator = match config.indicator.indicator_type {
             IndicatorType::Strings { strings } => {
-                let strings_layout = pango::Layout::new(&pango_context);
+                let indicator_layout = pango::Layout::new(&pango_context);
                 Indicator::Strings(indicator::Strings::new(
                     config.indicator.common,
                     strings,
-                    strings_layout,
+                    indicator_layout,
                     debug,
+                    text_height,
                 )?)
             }
             IndicatorType::Classic { classic } => Indicator::Classic(indicator::Classic::new(
@@ -899,6 +952,11 @@ impl Dialog {
             pango_context,
         };
 
+        debug!(
+            "layout: vertical_spacing: {}, horizontal_spacing: {}",
+            config.layout_opts.horizontal_spacing(text_height),
+            config.layout_opts.vertical_spacing(text_height)
+        );
         let (width, height) = config.layout_opts.layout.get_fn()(
             &config.layout_opts,
             &mut components,
@@ -920,50 +978,40 @@ impl Dialog {
             mouse_middle_pressed: false,
             background: config.background.into(),
             input_timeout_duration: config.input_timeout.map(Duration::from_secs),
+            input_timeout: None,
             debug,
-            cursor_size,
+            uses_cursor,
+            button_pressed: false,
         })
     }
 
-    pub fn paint_input_cursor(&self, cr: &cairo::Context, width: u16, height: u16) -> (u16, u16) {
-        // operator source required to init the picmap with alpha
-        cr.set_operator(cairo::Operator::Source);
-        cr.set_source_rgba(0.0, 0.0, 0.0, 0.0);
-        cr.paint().unwrap();
-
-        let w = width as f64;
-        cr.set_source_rgba(1.0, 1.0, 1.0, 1.0);
-        cr.set_line_width(3.0);
-        cr.move_to(0.0, 1.5);
-        cr.line_to(w, 1.5);
-        cr.move_to(0.0, height as f64 - 1.5);
-        cr.line_to(w, height as f64 - 1.5);
-        cr.move_to(w / 2.0, 0.0);
-        cr.line_to(w / 2.0, height as f64);
-        cr.stroke().unwrap();
-
-        cr.set_source_rgba(0.0, 0.0, 0.0, 1.0);
-        cr.set_line_width(1.0);
-        cr.move_to(1.0, 1.5);
-        cr.line_to(w - 1.0, 1.5);
-        cr.move_to(1.0, height as f64 - 1.5);
-        cr.line_to(w - 1.0, height as f64 - 1.5);
-        cr.move_to(w / 2.0, 1.0);
-        cr.line_to(w / 2.0, (height - 2) as f64);
-        cr.stroke().unwrap();
-
-        (width / 2, height / 2)
-    }
-
     pub fn on_displayed(&mut self, serial: FrameId) -> bool {
         self.indicator.on_displayed(serial)
     }
 
-    pub fn update(&mut self, cr: &cairo::Context, serial: FrameId) {
-        self.indicator.update(cr, &self.background);
+    pub fn set_painted(&mut self, serial: FrameId) {
         trace!("update serial {:?}", serial);
         self.indicator.set_painted(serial);
-        for (i, b) in self.buttons.iter_mut().enumerate() {
+        for b in &mut self.buttons {
+            b.set_painted();
+        }
+    }
+
+    pub fn dirty(&self) -> bool {
+        if self.indicator.dirty() {
+            return true;
+        }
+        for b in &self.buttons {
+            if b.dirty {
+                return true;
+            }
+        }
+        false
+    }
+
+    pub fn repaint(&self, cr: &cairo::Context) {
+        self.indicator.repaint(cr, &self.background);
+        for (i, b) in self.buttons.iter().enumerate() {
             if b.dirty {
                 trace!("button {} dirty", i);
                 b.clear(cr, &self.background);
@@ -983,151 +1031,62 @@ impl Dialog {
         // TODO can I preserve antialiasing without clearing the image first?
         cr.set_source(&self.background).unwrap();
         cr.paint().unwrap();
-        self.paint(cr, serial);
+        self.paint(cr);
+        self.set_painted(serial);
     }
 
-    fn paint(&mut self, cr: &cairo::Context, serial: FrameId) {
+    fn paint(&self, cr: &cairo::Context) {
         trace!("paint");
-        for l in &mut self.labels {
+        for l in &self.labels {
             l.paint(cr);
         }
         self.indicator.paint(cr);
-        self.indicator.set_painted(serial);
-        for b in &mut self.buttons {
+        for b in &self.buttons {
             b.paint(cr);
         }
     }
 
-    pub async fn run_events(mut self, xcontext: &mut XContext<'_>) -> Result<Option<Passphrase>> {
-        tokio::pin! { let xevents_ready = xcontext.conn.xfd.readable(); }
-
-        tokio::pin! { let input_timeout = sleep(self.input_timeout_duration.unwrap_or_else(||Duration::from_secs(0))); }
+    pub fn init_events(&mut self) {
         self.indicator.init_timeouts();
-        loop {
-            xcontext.conn.flush()?;
-            tokio::select! {
-                _ = &mut input_timeout, if self.input_timeout_duration.is_some() => {
-                    info!("input timeout");
-                    return Ok(None)
-                }
-                updated = self.indicator.handle_events(), if self.indicator.requests_events() => {
-                    if updated {
-                        xcontext.backbuffer.update(&mut self)?;
-                    }
-                }
-                xevents_guard = &mut xevents_ready => {
-                    match xcontext.poll_for_event()? {
-                        None => {
-                            xevents_guard.unwrap().clear_ready();
-                        }
-                        Some(event) => {
-                            match event {
-                                Event::Motion { x, y } => {
-                                    if self.handle_motion(x, y, xcontext)? {
-                                        xcontext.backbuffer.update(&mut self)?;
-                                    }
-                                }
-                                Event::KeyPress(key_press) => {
-                                    if let Some(timeout) = self.input_timeout_duration {
-                                        input_timeout
-                                            .as_mut()
-                                            .reset(Instant::now().checked_add(timeout).unwrap());
-                                    }
+        self.input_timeout = Some(Box::pin(sleep(
+            self.input_timeout_duration
+                .unwrap_or_else(|| Duration::from_secs(0)),
+        )));
+    }
 
-                                    let (action, dirty) = self.handle_key_press(key_press, xcontext)?;
-                                    trace!("action {:?}", action);
-                                    match action {
-                                        Action::Ok => return Ok(Some(self.indicator.into_pass())),
-                                        Action::Cancel => return Ok(None),
-                                        Action::Nothing => {},
-                                        _ => unreachable!(),
-                                    }
-                                    if dirty {
-                                        trace!("dirty");
-                                        xcontext.backbuffer.update(&mut self)?;
-                                    }
-                                }
-                                Event::ButtonPress { button, x, y, isrelease } => {
-                                    if let Some(timeout) = self.input_timeout_duration {
-                                        input_timeout
-                                            .as_mut()
-                                            .reset(Instant::now().checked_add(timeout).unwrap());
-                                    }
-
-                                    let (action, mut dirty) = self.handle_button_press(button, x, y, isrelease);
-                                    match action {
-                                        Action::Ok => return Ok(Some(self.indicator.into_pass())),
-                                        Action::Cancel => return Ok(None),
-                                        Action::PastePrimary => {
-                                            xcontext.paste_primary()?;
-                                        }
-                                        Action::PasteClipboard => {
-                                            xcontext.paste_clipboard()?;
-                                        }
-                                        Action::PlainText => {
-                                            self.indicator.toggle_plaintext();
-                                            self.buttons[3].toggle();
-                                            dirty = true;
-                                        }
-                                        _ => {}
-                                    }
-                                    if dirty {
-                                        xcontext.backbuffer.update(&mut self)?;
-                                    }
-                                }
-                                Event::Paste(mut val) => {
-                                    if self.indicator.pass_insert(&val, true) {
-                                        xcontext.backbuffer.update(&mut self)?;
-                                    }
-                                    val.zeroize();
-                                }
-                                Event::Focus(focus) => {
-                                    if self.indicator.set_focused(focus) {
-                                        xcontext.backbuffer.update(&mut self)?;
-                                    }
-                                }
-                                Event::PendingUpdate => {
-                                    xcontext.backbuffer.update(&mut self)?;
-                                }
-                                Event::VsyncCompleted(serial) => {
-                                    if self.on_displayed(serial) {
-                                        trace!("on displayed dirty");
-                                        xcontext.backbuffer.update(&mut self)?;
-                                    }
-                                }
-                                Event::Exit => {
-                                    return Ok(None);
-                                }
-                            }
-                        }
-                    }
-                    tokio::task::yield_now().await;
-                    xevents_ready.set(xcontext.conn.xfd.readable());
-                }
+    pub async fn handle_events(&mut self) -> Action {
+        tokio::select! {
+            _ = self.input_timeout.as_mut().unwrap(), if self.input_timeout_duration.is_some() => {
+                info!("input timeout");
+                Action::Cancel
+            }
+            _ = self.indicator.handle_events(), if self.indicator.requests_events() => {
+                Action::Nothing
             }
         }
     }
 
-    pub fn handle_motion(&mut self, x: f64, y: f64, xcontext: &XContext) -> Result<bool> {
+    pub fn handle_motion(&mut self, x: f64, y: f64, xcontext: &XContext) -> Result<()> {
         let mut found = false;
-        let mut dirty = false;
         for b in &mut self.buttons {
             if found {
+                trace!("set_hover: false");
                 b.set_hover(false);
             } else if b.is_inside(x, y) {
-                b.set_hover(true);
+                trace!("set_hover: {}", self.button_pressed == b.pressed);
+                b.set_hover(self.button_pressed == b.pressed);
                 found = true;
             } else {
+                trace!("set_hover: false");
                 b.set_hover(false);
             }
-            dirty = dirty || b.dirty;
         }
         if !found && self.indicator.is_inside(x, y) {
             self.indicator.set_hover(true, xcontext)?;
         } else {
             self.indicator.set_hover(false, xcontext)?;
         };
-        Ok(dirty)
+        Ok(())
     }
 
     fn cairo_context_changed(&mut self, cr: &cairo::Context) {
@@ -1139,16 +1098,10 @@ impl Dialog {
         }
     }
 
-    pub fn resize(
-        &mut self,
-        cr: &cairo::Context,
-        width: u16,
-        height: u16,
-        serial: FrameId,
-        surface_cleared: bool,
-    ) {
+    pub fn resize(&mut self, cr: &cairo::Context, width: u16, height: u16, surface_cleared: bool) {
         cr.set_source(&self.background).unwrap();
 
+        // TODO put to clear()
         if surface_cleared {
             // clear the whole buffer
             cr.paint().unwrap();
@@ -1185,7 +1138,7 @@ impl Dialog {
 
         self.cairo_context_changed(cr);
 
-        self.paint(cr, serial);
+        self.paint(cr);
     }
 
     pub fn handle_button_press(
@@ -1194,53 +1147,81 @@ impl Dialog {
         x: f64,
         y: f64,
         isrelease: bool,
-    ) -> (Action, bool) {
-        if !isrelease && button == xproto::ButtonIndex::M2 {
+        xcontext: &mut XContext,
+    ) -> Result<Action> {
+        if let Some(timeout) = self.input_timeout_duration {
+            self.input_timeout
+                .as_mut()
+                .unwrap()
+                .as_mut()
+                .reset(Instant::now().checked_add(timeout).unwrap());
+        }
+
+        let action = if !isrelease && button == xproto::ButtonIndex::M2 {
             self.mouse_middle_pressed = true;
+            Action::Nothing
         } else if self.mouse_middle_pressed && button == xproto::ButtonIndex::M2 {
             self.mouse_middle_pressed = false;
             if x >= 0.0 && x < self.width as f64 && y >= 0.0 && y < self.height as f64 {
-                trace!("PRIMARY selection");
-                return (Action::PastePrimary, false);
+                Action::PastePrimary
+            } else {
+                Action::Nothing
             }
         } else if button != xproto::ButtonIndex::M1 {
             trace!("not the left mouse button");
+            Action::Nothing
         } else {
-            return self.handle_mouse_left_button_press(x, y, isrelease);
+            self.handle_mouse_left_button_press(x, y, isrelease)
+        };
+
+        match action {
+            Action::Ok => return Ok(Action::Ok),
+            Action::Cancel => return Ok(Action::Cancel),
+            Action::PastePrimary => {
+                xcontext.paste_primary()?;
+            }
+            Action::PasteClipboard => {
+                xcontext.paste_clipboard()?;
+            }
+            Action::PlainText => {
+                self.indicator.toggle_plaintext();
+                self.buttons[3].toggle();
+            }
+            _ => {}
         }
-        (Action::Nothing, false)
+
+        Ok(Action::Nothing)
     }
 
     // Return true iff dialog should be repainted
-    fn handle_mouse_left_button_press(&mut self, x: f64, y: f64, release: bool) -> (Action, bool) {
-        let mut dirty = false;
+    fn handle_mouse_left_button_press(&mut self, x: f64, y: f64, release: bool) -> Action {
         if release {
+            self.button_pressed = false;
             for (i, b) in self.buttons.iter_mut().enumerate() {
                 if b.pressed {
                     b.set_pressed(false);
                     if b.is_inside(x, y) {
                         trace!("release inside button {}", i);
-                        return (Components::ACTIONS[i], b.dirty);
-                    } else {
-                        return (Action::Nothing, b.dirty);
+                        return Components::ACTIONS[i];
                     }
+                    return Action::Nothing;
                 }
             }
         } else {
-            let (inside, d) = self.indicator.set_cursor(x, y);
-            dirty = dirty || d;
+            let inside = self.indicator.set_cursor(x, y);
             if inside {
-                return (Action::Nothing, dirty);
+                return Action::Nothing;
             }
             for (i, b) in self.buttons.iter_mut().enumerate() {
                 if b.is_inside(x, y) {
                     trace!("inside button {}", i);
                     b.set_pressed(true);
-                    return (Action::Nothing, dirty || b.dirty);
+                    self.button_pressed = true;
+                    return Action::Nothing;
                 }
             }
         }
-        (Action::Nothing, dirty)
+        Action::Nothing
     }
 
     fn get_secure_utf8_do(
@@ -1278,7 +1259,15 @@ impl Dialog {
         &mut self,
         key_press: Keypress,
         xcontext: &mut XContext,
-    ) -> Result<(Action, bool)> {
+    ) -> Result<Action> {
+        if let Some(timeout) = self.input_timeout_duration {
+            self.input_timeout
+                .as_mut()
+                .unwrap()
+                .as_mut()
+                .reset(Instant::now().checked_add(timeout).unwrap());
+        }
+
         let keyboard = &xcontext.keyboard;
         let key = key_press.get_key();
         let mut key_sym = keyboard.key_get_one_sym(key);
@@ -1292,7 +1281,7 @@ impl Dialog {
                 match compose.state_get_status() {
                     xkb_compose_status::XKB_COMPOSE_NOTHING => {}
                     xkb_compose_status::XKB_COMPOSE_COMPOSING => {
-                        return Ok((Action::Nothing, false));
+                        return Ok(Action::Nothing);
                     }
                     xkb_compose_status::XKB_COMPOSE_COMPOSED => {
                         key_sym = compose.state_get_one_sym();
@@ -1300,7 +1289,7 @@ impl Dialog {
                     }
                     xkb_compose_status::XKB_COMPOSE_CANCELLED => {
                         compose.state_reset();
-                        return Ok((Action::Nothing, false));
+                        return Ok(Action::Nothing);
                     }
                     _ => unreachable!(),
                 }
@@ -1314,25 +1303,21 @@ impl Dialog {
 
         let mut matched = true;
         let mut action = Action::Nothing;
-        let dirty = match key_sym {
+        match key_sym {
             keysyms::XKB_KEY_Return | keysyms::XKB_KEY_KP_Enter => {
                 action = Action::Ok;
-                false
             }
             keysyms::XKB_KEY_j | keysyms::XKB_KEY_m if ctrl => {
                 action = Action::Ok;
-                false
             }
             keysyms::XKB_KEY_Escape => {
                 action = Action::Cancel;
-                false
             }
             keysyms::XKB_KEY_BackSpace => self.indicator.pass_delete(),
             keysyms::XKB_KEY_h if ctrl => self.indicator.pass_delete(),
             keysyms::XKB_KEY_u if ctrl => self.indicator.pass_clear(),
             keysyms::XKB_KEY_v if ctrl => {
                 xcontext.paste_clipboard()?;
-                false
             }
             keysyms::XKB_KEY_Left => self.indicator.move_cursor(-1),
             keysyms::XKB_KEY_Right => self.indicator.move_cursor(1),
@@ -1343,24 +1328,22 @@ impl Dialog {
                 ) =>
             {
                 xcontext.paste_primary()?;
-                false
             }
             _ => {
                 matched = false;
-                false
             }
         };
         key_sym.zeroize();
         if matched {
-            return Ok((action, dirty));
+            return Ok(action);
         }
 
         let buf = self.get_secure_utf8_do(&xcontext.keyboard, key, composed);
         let s = unsafe { std::str::from_utf8_unchecked(buf.unsecure()) };
         if !s.is_empty() {
-            let dirty = self.indicator.pass_insert(s, false);
-            return Ok((Action::Nothing, dirty));
+            self.indicator.pass_insert(s, false);
+            return Ok(Action::Nothing);
         }
-        Ok((Action::Nothing, dirty))
+        Ok(Action::Nothing)
     }
 }
