@@ -9,7 +9,6 @@ use tokio::io::unix::AsyncFd;
 use tokio::signal::unix::{signal, SignalKind};
 use tokio::time::Instant;
 use x11rb::connection::{Connection as _, RequestConnection as _};
-use x11rb::protocol::render;
 use x11rb::protocol::xproto::{self, ConnectionExt as _};
 use x11rb::{atom_manager, properties};
 // for change_propertyN()
@@ -24,7 +23,7 @@ mod event;
 mod keyboard;
 mod secret;
 
-use errors::{anyhow, Context as _, Result, X11ErrorString as _};
+use errors::{anyhow, Context as _, Result};
 use secret::Passphrase;
 
 pub const CLASS: &str = "SshAskpass";
@@ -37,10 +36,11 @@ atom_manager! {
         WM_PROTOCOLS,
         WM_DELETE_WINDOW,
         _NET_WM_ICON,
+        _NET_WM_ICON_NAME,
         _NET_WM_NAME,
         _NET_WM_PID,
         _NET_WM_WINDOW_TYPE,
-        _NET_WM_WINDOW_TYPE_NORMAL,
+        _NET_WM_WINDOW_TYPE_DIALOG,
         _NET_WM_STATE,
         _NET_WM_STATE_ABOVE,
         UTF8_STRING,
@@ -76,67 +76,6 @@ impl Deref for Connection {
     }
 }
 
-fn create_input_cursor(
-    conn: &Connection,
-    screen: &xproto::Screen,
-    dialog: &dialog::Dialog,
-    window: xproto::Window,
-) -> Result<XId> {
-    //let render_version = render::query_version(conn.xfd.get_ref(), 0, 8)?.reply().map_xerr(conn);
-    let pict_format = render::query_pict_formats(conn.deref())?
-        .reply()
-        .map_xerr()?
-        .formats
-        .iter()
-        .find_map(|format| {
-            if format.type_ == render::PictType::DIRECT
-                && format.depth == 32
-                && format.direct.red_shift == 16
-                && format.direct.red_mask == 0xff
-                && format.direct.green_shift == 8
-                && format.direct.green_mask == 0xff
-                && format.direct.blue_shift == 0
-                && format.direct.blue_mask == 0xff
-                && format.direct.alpha_shift == 24
-                && format.direct.alpha_mask == 0xff
-            {
-                Some(format.id)
-            } else {
-                None
-            }
-        })
-        .expect("The X11 server is missing the RENDER ARGB_32 standard format!");
-
-    let depth_type = screen
-        .allowed_depths
-        .iter()
-        .find(|d| d.depth == 32)
-        .expect("invalid depth");
-    let visual_type = depth_type
-        .visuals
-        .get(0)
-        .expect("depth has no visual types");
-
-    let width: u16 = dialog.cursor_size.unwrap().0.try_into().unwrap();
-    let height: u16 = dialog.cursor_size.unwrap().1.try_into().unwrap();
-    let surface = backbuffer::XcbSurface::new(conn, window, 32, visual_type, width, height)?;
-    let cr = cairo::Context::new(&surface).expect("cairo context new");
-    let (hot_x, hot_y) = dialog.paint_input_cursor(&cr, width, height);
-    surface.flush();
-    let picture = conn.generate_id().map_xerr()?;
-    render::create_picture(
-        conn.deref(),
-        picture,
-        surface.pixmap,
-        pict_format,
-        &Default::default(),
-    )?;
-    let cursor = conn.generate_id().map_xerr()?;
-    render::create_cursor(conn.deref(), cursor, picture, hot_x, hot_y)?;
-    render::free_picture(conn.deref(), picture)?;
-    Ok(cursor)
-}
-
 async fn run_xcontext(
     cfg_loader: config::Loader,
     opts: Opts,
@@ -203,7 +142,7 @@ async fn run_xcontext(
         screen.default_colormap
     } else {
         debug!("depth requires a new colormap");
-        let colormap = conn.generate_id().map_xerr()?;
+        let colormap = conn.generate_id()?;
         conn.create_colormap(
             xproto::ColormapAlloc::NONE,
             colormap,
@@ -213,7 +152,7 @@ async fn run_xcontext(
         colormap
     };
 
-    let window = conn.generate_id().map_xerr()?;
+    let window = conn.generate_id()?;
     conn.create_window(
         depth,
         window,
@@ -240,7 +179,7 @@ async fn run_xcontext(
             .colormap(colormap),
     )?;
 
-    let atoms = atoms.reply().map_xerr()?;
+    let atoms = atoms.reply()?;
 
     let hostname = if let Some(hn) = std::env::var_os("HOSTNAME") {
         hn
@@ -257,6 +196,13 @@ async fn run_xcontext(
         xproto::AtomEnum::STRING,
         title.as_bytes(),
     )?;
+    conn.change_property8(
+        xproto::PropMode::REPLACE,
+        window,
+        xproto::AtomEnum::WM_ICON_NAME,
+        xproto::AtomEnum::STRING,
+        title.as_bytes(),
+    )?;
     conn.change_property8(
         xproto::PropMode::REPLACE,
         window,
@@ -264,6 +210,13 @@ async fn run_xcontext(
         atoms.UTF8_STRING,
         title.as_bytes(),
     )?;
+    conn.change_property8(
+        xproto::PropMode::REPLACE,
+        window,
+        atoms._NET_WM_ICON_NAME,
+        atoms.UTF8_STRING,
+        title.as_bytes(),
+    )?;
     conn.change_property8(
         xproto::PropMode::REPLACE,
         window,
@@ -292,7 +245,7 @@ async fn run_xcontext(
         window,
         atoms._NET_WM_WINDOW_TYPE,
         xproto::AtomEnum::ATOM,
-        &[atoms._NET_WM_WINDOW_TYPE_NORMAL],
+        &[atoms._NET_WM_WINDOW_TYPE_DIALOG],
     )?;
     // be above of other windows
     conn.change_property32(
@@ -336,17 +289,18 @@ async fn run_xcontext(
         )?;
     }
 
-    // try to prevent resizing
-    let size_hints = properties::WmSizeHints {
+    let mut size_hints = properties::WmSizeHints {
         size: Some((
             properties::WmSizeHintsSpecification::ProgramSpecified,
             window_width.into(),
             window_height.into(),
         )),
         min_size: Some((window_width.into(), window_height.into())),
-        max_size: Some((window_width.into(), window_height.into())),
         ..properties::WmSizeHints::default()
     };
+    if !config.resizable {
+        size_hints.max_size = Some((window_width.into(), window_height.into()));
+    }
     size_hints.set_normal_hints(&*conn, window)?;
 
     debug!("map window");
@@ -355,6 +309,19 @@ async fn run_xcontext(
     conn.flush()?;
 
     // Load the slow ones after we have mapped the window
+
+    let resource_db;
+    let cursor_handle = if dialog.uses_cursor {
+        resource_db = x11rb::resource_manager::Database::new_from_default(&*conn)?;
+        Some(x11rb::cursor::Handle::new(
+            &*conn,
+            screen_num,
+            &resource_db,
+        )?)
+    } else {
+        None
+    };
+
     debug!("dialog init");
     let mut backbuffer = backbuffer.reply()?;
     backbuffer.init(window, &mut dialog)?;
@@ -362,19 +329,16 @@ async fn run_xcontext(
     debug!("keyboard init");
     let keyboard = keyboard::Keyboard::new(&conn)?;
 
-    let input_cursor = if dialog.cursor_size.is_some()
-        && conn
-            .extension_information(render::X11_EXTENSION_NAME)?
-            .is_some()
-    {
-        debug!("cursor init");
-        Some(create_input_cursor(&conn, screen, &dialog, window)?)
+    debug!("cursor init");
+    let input_cursor = if let Some(cursor_handle) = cursor_handle {
+        let cursor_handle = cursor_handle.reply()?;
+        Some(cursor_handle.load_cursor(&*conn, "xterm").unwrap())
     } else {
         None
     };
     debug!("init took {}ms", startup_time.elapsed().as_millis());
 
-    let mut xcon = event::XContext {
+    let mut xcontext = event::XContext {
         keyboard,
         conn: &conn,
         backbuffer,
@@ -391,7 +355,7 @@ async fn run_xcontext(
         input_cursor,
     };
 
-    dialog.run_events(&mut xcon).await
+    xcontext.run_events(dialog).await
 }
 
 #[derive(Clap)]
@@ -411,7 +375,7 @@ struct Opts {
     quiet: bool,
 
     #[clap(short, long, parse(from_occurrences))]
-    /// Increases the level of verbosity (the max level is -vvvv)
+    /// Increases the level of verbosity (the max level is -vv)
     verbose: usize,
 
     #[clap(short, long)]
@@ -425,10 +389,9 @@ struct Opts {
     /// Label in the dialog.
     label: Option<String>,
 
-    /// Generate default config into path. Consider using the supplied default
-    /// configuration file with comments instead.
-    #[clap(long, hidden = true)]
-    gen_config: Option<PathBuf>,
+    /// Output default config to stdout.
+    #[clap(long)]
+    gen_config: bool,
 }
 
 fn run() -> i32 {
@@ -445,7 +408,7 @@ fn run() -> i32 {
     let opts = Opts::from_arg_matches(&app.get_matches()).expect("from_arg_matches");
 
     let mut log = stderrlog::new();
-    log.quiet(opts.quiet).verbosity(opts.verbose);
+    log.quiet(opts.quiet).verbosity(opts.verbose + 2);
     if opts.debug {
         log.timestamp(stderrlog::Timestamp::Millisecond);
     }
@@ -454,16 +417,16 @@ fn run() -> i32 {
     match run_logged(cfg_loader, opts, startup_time) {
         Ok(ret) => ret,
         Err(err) => {
-            error!("{:#}", err);
+            error!("{}", err);
             2
         }
     }
 }
 
 fn run_logged(cfg_loader: config::Loader, opts: Opts, startup_time: Instant) -> Result<i32> {
-    if let Some(path) = opts.gen_config {
+    if opts.gen_config {
         let cfg = config::Config::default();
-        cfg_loader.save_path(&path, &cfg)?;
+        cfg_loader.print(&cfg)?;
         return Ok(0);
     }
 
